/* Copyright (C) 2002, 2003, 2004 Shell Technologies s.r.l.
   Authors: Marco Cesati <cesati@uniroma2.it>
   
   driver for the Galil 1800 PCI motion controllers (http://www.galilmc.com/)
   
   loosely inspired by Galil's own Linux driver, as well as by a bunch of
   PCI drivers in the kernel source code
   
   This program is free software; you can redistribute  it and/or modify it
   under  the terms of  the GNU General  Public License as published by the
   Free Software Foundation;  either version 2 of the  License, or (at your
   option) any later version.
   
   Programmers list: 
   
   		MC: Marco Cesati
        
   Revision history:
   
     8 November 2002 - ver. 0.1 (MC): first release (kernel 2.4.19)
    12 December 2002 - ver. 0.2 (MC): support for secondary communication
                                             channel (kernel 2.4.20)
    13 December 2002 - ver. 0.3 (MC): ioctl() command to fetch interrupt
                                             stack
    30 January 2003  - ver. 0.4 (MC): fixed race condition in 
                                             galil_1800_send_signals()
    14 July 2004    -  ver. 0.4.1 (MC): fixed silly bug when looking for the
                                             minor number in galil_1800_open()
                                        avoid endless loop in galil_1800_get_serial_number()
*/

#include <linux/version.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/spinlock.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/hardirq.h>

#include "galil_ioctl.h"

#define GALIL_1800_MODULE_NAME "Galil 1800"
#define PFX GALIL_1800_MODULE_NAME ": "

#define GALIL_1800_READ_TIMEOUT      10 /* ticks (1 tick = ~10ms) */
#define GALIL_1800_WRITE_TIMEOUT     10 /* ticks (1 tick = ~10ms) */

#define PLX_PCI_VENDOR_ID            0x10B5
#define PLX_PCI_DEVICE_ID            0x9050
#define GALIL_SUB_VENDOR             0x1079
#define GALIL_SUB_DEVICE_1800        0x1800

#define GALIL_1800_READ_FIFO_SIZE    256
#define GALIL_1800_WRITE_FIFO_SIZE   256
#define GALIL_1800_READ_FIFO2_SIZE   256


/* main device driver descriptor, one for every registered device */

struct galil_1800_state {
    struct list_head devs;       /* list of Galil 1800 devices */
    struct pci_dev *dev;         /* the PCI main device descriptor */
    unsigned long io;            /* base I/O address */
    unsigned long ctrl_reg;      /* control register I/O address */
    unsigned long irq_reg;       /* irq register I/O address */
    unsigned long chan2_reg;     /* secondary channel register I/O address */
    unsigned int irq;            /* IRQ line number */
    int index;                   /* index (== minor number) of the device */
    int opens;                   /* how many processes have opened the device
                                    file; protected by open_sem */
    struct semaphore open_sem;   /* semaphore for concurrent open requests */
    mode_t open_mode;            /* the read/write mode of the device file */
    wait_queue_head_t open_wait; /* waiting for an open to complete */
    spinlock_t hwlock;           /* spin lock for low-level fiddling */
    rwlock_t siglock;            /* r/w spin lock for the list of processes */
    unsigned char irqstack[GALIL_1800_IRQ_STACK_SZ]; /* stack of IRQ values */
    int irqstack_sz;             /* current number of values in the stack */
    spinlock_t irqlock;          /* spin lock for the interrupt handler */
    struct list_head siglist;    /* processes to be signaled on interrupts */
    struct tasklet_struct tlet;  /* tasklet handling the siglist list of 
                                    processes to be signaled */
    unsigned long serial;        /* the serial number read from the device */
};

/* element of the list of processes to be signaled on interrupt occurrences */

struct galil_1800_siglist_t {
    struct list_head list;    /* pointers for the next and previous elements */
    struct task_struct * owner;  /* pointer to the process descriptor */
    pid_t pid;               /* pid of the process descriptor (safety check) */
    char *ubuf;                  /* buffer in the User Mode space for the interrupt values */
};

#define GALIL_1800_READ_REG(A)         ((A)->io)
#define GALIL_1800_WRITE_REG(A)        GALIL_1800_READ_REG(A)
#define GALIL_1800_CONTROL_REG(A)      ((A)->ctrl_reg)
#define GALIL_1800_IRQ_REG(A)          ((A)->irq_reg)
#define GALIL_1800_RESET_REG(A)        GALIL_1800_IRQ_REG(A)
#define GALIL_1800_2CHAN_REG(A)        ((A)->chan2_reg)

#define GALIL_1800_FIFO_FULL_FLAG      (0x01)
#define GALIL_1800_FIFO_HALF_FLAG      (0x02)
#define GALIL_1800_FIFO_EMPTY_FLAG     (0x04)
#define GALIL_1800_FIFO2_BUSY_FLAG     (0x08)
#define GALIL_1800_FIFO2_FREEZE_FLAG   (0x10)
#define GALIL_1800_IRQ_STATUS_FLAG     (0x20)
#define GALIL_1800_IRQ_ENABLE_FLAG     (0x40)
#define GALIL_1800_FIFO2_EMPTY_FLAG    (0x80)

/* the list of galil_1800_state descriptors */
static LIST_HEAD(galil_1800_devs);

/* a read/write spin lock that protects the above list in MP */
static rwlock_t galil_1800_devs_lock = RW_LOCK_UNLOCKED;

/* index (minor number) of the next device yet to be registered */
static unsigned int devindex = 0;

/* major number of associated device files */
static int major = -1;


/* 
   readwrite == 0: clear both read and write FIFO
   readwrite == 1: clear read FIFO only
   readwrite == 2: clear write FIFO only
*/

static inline void __galil_1800_clear_FIFO(struct galil_1800_state *dev, 
                                           int readwrite)
{
    char mask = (!readwrite ? 6 : (readwrite == 1 ? 2 : 4));
    outb(mask, GALIL_1800_RESET_REG(dev));    
}

static inline void galil_1800_clear_FIFO(struct galil_1800_state *dev, 
                                         int readwrite)
{
    unsigned long flags;
    spin_lock_irqsave(&dev->hwlock, flags);
    __galil_1800_clear_FIFO(dev, readwrite);
    spin_unlock_irqrestore(&dev->hwlock, flags);
}

static inline void galil_1800_enable_IRQ(struct galil_1800_state *dev)
{
    char t;
    unsigned long flags;
    
    spin_lock_irqsave(&dev->hwlock, flags);
     (void) inb(GALIL_1800_IRQ_REG(dev)); /* useless ?! */
    t = inb(GALIL_1800_CONTROL_REG(dev));
     outb(t|GALIL_1800_IRQ_ENABLE_FLAG, GALIL_1800_CONTROL_REG(dev));
     (void) inb(GALIL_1800_IRQ_REG(dev)); /* useless ?! */
    spin_unlock_irqrestore(&dev->hwlock, flags);
}

static inline void galil_1800_disable_IRQ(struct galil_1800_state *dev)
{
    unsigned long flags;
    char t;
    
    spin_lock_irqsave(&dev->hwlock, flags);
    t = inb(GALIL_1800_CONTROL_REG(dev)) & ~GALIL_1800_IRQ_ENABLE_FLAG;
    outb(t, GALIL_1800_CONTROL_REG(dev));
    spin_unlock_irqrestore(&dev->hwlock, flags);
}

/* Wait for a byte in the Read FIFO, or until a timeout occurs.
   Return 1 if a byte is available in the FIFO, 0 otherwise. 
   Interrupts MUST be enabled, or the kernel will freeze */

static inline int galil_1800_may_read(struct galil_1800_state *dev)
{
    unsigned long timeout = jiffies + GALIL_1800_READ_TIMEOUT;
    
    do {
        char t = inb(GALIL_1800_CONTROL_REG(dev)) & GALIL_1800_FIFO_EMPTY_FLAG;
        if (!t)
            return 1;
    } while (jiffies < timeout);
    return 0;
}

/* Wait for a free slot in the Write FIFO, or until a timeout occurs.
   Return 2 if FIFO is less than half full (at least 128 free slots),
   1 if a slot is available in the FIFO, 0 otherwise.
   Interrupts MUST be enabled, or the kernel will freeze */

static inline int galil_1800_may_write(struct galil_1800_state *dev)
{
    unsigned long timeout = jiffies + GALIL_1800_WRITE_TIMEOUT;
    
    do {
        char t = inb(GALIL_1800_CONTROL_REG(dev));
        if (!(t & GALIL_1800_FIFO_HALF_FLAG))
            return 2;
        if (!(t & GALIL_1800_FIFO_FULL_FLAG))
            return 1;
    } while (jiffies < timeout);
    return 0;    
}

/* Secondary Communication Channel: 

    - on DMC-1800, SCC can be used in polling FIFO mode only: data goes in
      the secondary FIFO of the controller, and then transferred to the UM
      address space by the driver
    - SCC is not active by default, it must be enabled by using the DR command
      with a negative argument -n (2^n samples between updates)
    - DR0 command disables SCC
    - SCC provide information at a fixed rate about position, position error,
      auxiliary position, velocity, torque, axis status, input line status,
      output line status, segment count for coordinate moves, and controller
      status
    - official docs are particularly obscure about SCC in 1800 devices, as 
      they fail in stating the differences between 1800 and 1802 devices;
      some information has been gathered from the buggy Galil original driver
    - our handshake procedure for accessing the secondary FIFO:
        1. "wait for data": read GALIL_1800_FIFO2_EMPTY_FLAG bit in 
           GALIL_1800_CONTROL_REG until it is equal to 0
        2. "freeze 2nd FIFO": set GALIL_1800_FIFO2_FREEZE_FLAG bit in
           GALIL_1800_CONTROL_REG
        3. "wait for complete": read GALIL_1800_FIFO2_BUSY_FLAG bit in 
           GALIL_1800_CONTROL_REG until it is equal to 0
        4. "fetch data": read byte at GALIL_1800_2CHAN_REG
        5. "check for data": read GALIL_1800_FIFO2_EMPTY_FLAG bit in 
           GALIL_1800_CONTROL_REG; if it is equal to 0, goto step 4.
        6. "unfreeze 2nd FIFO": clear GALIL_1800_FIFO2_FREEZE_FLAG bit in
           GALIL_1800_CONTROL_REG
      that's different from the procedure in the Galil original driver, and it's
      different from what's described in the official docs, but it has a slightly
      better chance to work, IMHO
*/
    
/* Wait for a byte in the secondary FIFO, or until a timeout occurs.
   Return 1 if a byte is available in the 2nd FIFO, 0 otherwise. 
   Interrupts MUST be enabled, or the kernel will freeze */

static inline int galil_1800_may_read_FIFO2(struct galil_1800_state *dev)
{
    unsigned long timeout = jiffies + GALIL_1800_READ_TIMEOUT;
    
    do {
        char t = inb(GALIL_1800_CONTROL_REG(dev)) & GALIL_1800_FIFO2_EMPTY_FLAG;
        if (!t)
            return 1;
    } while (jiffies < timeout);
    return 0;
}

/* Freeze and unfreeze the secondary communication channel */

static inline void galil_1800_freeze_FIFO2(struct galil_1800_state *dev)
{
 	char t = inb(GALIL_1800_CONTROL_REG(dev));
    t |= GALIL_1800_FIFO2_FREEZE_FLAG;
    outb(t, GALIL_1800_CONTROL_REG(dev));   
}

static inline void galil_1800_unfreeze_FIFO2(struct galil_1800_state *dev)
{
 	char t = inb(GALIL_1800_CONTROL_REG(dev));
    t &= ~GALIL_1800_FIFO2_FREEZE_FLAG;
    outb(t, GALIL_1800_CONTROL_REG(dev));
}

/* Wait until the controller ends transferring a data in the 2nd FIFO */

static inline void galil_1800_wait_transfer_FIFO2(struct galil_1800_state *dev)
{
	char t;
    
    do {
        t = inb(GALIL_1800_CONTROL_REG(dev)) & GALIL_1800_FIFO2_BUSY_FLAG;
    } while (t);    
}

/* Read from the secondary communication channel (2nd FIFO)
   - data are written in %buf%, a preallocated buffer of size %max%
   - return the number of bytes read (even 0) */

static inline int galil_1800_read_FIFO2(struct galil_1800_state *dev, 
                                        char *buf, int max)
{
	char *p = buf;
    char t;
    int n = 0;
	
    if (!galil_1800_may_read_FIFO2(dev))
        return n;
	galil_1800_freeze_FIFO2(dev);
	galil_1800_wait_transfer_FIFO2(dev);
	do {
        t = inb(GALIL_1800_2CHAN_REG(dev));
        *p++ = t;
        ++n;
        t = inb(GALIL_1800_CONTROL_REG(dev)) & GALIL_1800_FIFO2_EMPTY_FLAG;
    } while (!t && n<max);
    galil_1800_unfreeze_FIFO2(dev);
	return n;                
}

/* This feature is undocumented, info from the Galil's Linux driver. 
   It appears that they would use the S/N of their cards as a "private key"
   known only to licensed user. Of course, this is yet another example of the
   "security by hiding" approach, which is very weak and useless for any
   practical purpose. Anyway, Galil has published the source code of its 
   Linux driver... */

/* Return <0 in case of errors, 0 otherwise. 
  The serial number is written into %buf% */
 
static int galil_1800_get_serial_number(struct galil_1800_state *dev, 
                                        char *buf, int len)
{
    unsigned long flags;
    int t, i;
    unsigned long tot = 0;
    char *p = buf;
    
    if (!buf)
        return -EINVAL;
        
    galil_1800_clear_FIFO(dev, 0);
    
    t = galil_1800_may_write(dev);
    
    if (t == 0 || t == 1) {
        printk(KERN_ERR PFX "cannot flush the Write FIFO (code=%d) \n", t);
        return -ENOSPC;
    }

    spin_lock_irqsave(&dev->hwlock, flags);    
    outb( 'M',  GALIL_1800_WRITE_REG(dev) );
    outb( 'G',  GALIL_1800_WRITE_REG(dev) );
    outb( '_',  GALIL_1800_WRITE_REG(dev) );
    outb( 'B',  GALIL_1800_WRITE_REG(dev) );
    outb( 'N',  GALIL_1800_WRITE_REG(dev) );
    outb( '\r', GALIL_1800_WRITE_REG(dev) );
    spin_unlock_irqrestore(&dev->hwlock, flags);
    
    i=0;
    t=0;
    while (i < len) {
        char c;
        
        if (!galil_1800_may_read(dev)) {
            if (++t > 10)
                break;
            printk(KERN_INFO PFX "waiting for serial number on Read FIFO\n");
            schedule();
            continue;
        }
        c = inb(GALIL_1800_READ_REG(dev));
        if (c == '.' || c == 0x0d)
            break;
        if (c >= '0' && c <= '9')
            tot = tot * 10 + c - '0';
        *p++ = c;
        ++i;
        t = 0;
    }
    
    if (!i) {
    	printk(KERN_INFO PFX "device #%d didn't return its serial number\n", 
                             dev->index);
        return -1;   
    }

    *p = '\0';
    
    if (dev->serial != tot) {
        printk(KERN_INFO PFX "device #%d has serial number %lu (\"%s\")\n", 
                dev->index, tot, buf);
        dev->serial = tot;
    }
    return 0;
}


/* This is the interrupt handler.
   It sends a signal to any process in the siglist.
   Actually, this is done later in a deferrable function 
    (galil_1800_send_signals()). */

static void galil_1800_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    struct galil_1800_state *dev = dev_id;
    unsigned long flags;
    unsigned char status, t;
    int i;
    
    spin_lock_irqsave(&dev->hwlock, flags);

#if 0
    /* tech docs are not straightforward on how to get the interrupt status;
     in particular, should we write 0x06 in GALIL_1800_IRQ_REG before 
     reading the status? */
    
    outb(0x06, GALIL_1800_IRQ_REG(dev));
#endif
    
    status = inb(GALIL_1800_IRQ_REG(dev));
    t = inb(GALIL_1800_CONTROL_REG(dev));
    outb(t | GALIL_1800_IRQ_STATUS_FLAG, GALIL_1800_CONTROL_REG(dev));
    spin_unlock_irqrestore(&dev->hwlock, flags);
    
    if (!(t & GALIL_1800_IRQ_STATUS_FLAG))
        return;    /* shared interrupt raised by someone else */
    
    
    /* save the status value in the device descriptor's stack */

    i = dev->irqstack_sz++;
    if (i < GALIL_1800_IRQ_STACK_SZ)
        dev->irqstack[i] = status;
    else {
         printk(KERN_INFO PFX
               "IRQ stack is full, discarding value...\n");
        dev->irqstack_sz = GALIL_1800_IRQ_STACK_SZ;
    }

    /* schedule execution of galil_1800_send_signals() */
    
    tasklet_schedule(&dev->tlet);
}

static void galil_1800_send_signals(unsigned long device)
{
    struct galil_1800_state * dev = (struct galil_1800_state *) device;
    struct siginfo sinfo;
    struct list_head *list;
    struct list_head *head = &dev->siglist;
    struct galil_1800_siglist_t *elem;
    unsigned long flags;
    int sz, i;
    
    sinfo.si_signo = SIGUSR1;
    sinfo.si_errno = 0;

    sz = dev->irqstack_sz;    /* should irqstack_sz be atomic_t ? */

    read_lock(&dev->siglock); /* it's safe to leave interrupts enabled, 
                                 because no asynchronous function
                                 (interrupt handler or deferrable 
                                 function) gets the write siglock */
    list_for_each(list, head) {
         elem = list_entry(list, struct galil_1800_siglist_t, list);
         if (copy_to_user(elem->ubuf, dev->irqstack, sz))
             printk(KERN_ERR PFX "error while copying irq values, PID=%d buffer=%p\n",
                elem->pid, elem->ubuf);
         sinfo.si_code = sz;
         send_sig_info(SIGUSR1, &sinfo, elem->owner);
    }   
    read_unlock(&dev->siglock);

    spin_lock_irqsave(&dev->irqlock, flags);
    if (sz < dev->irqstack_sz)
    	for (i=0; i<dev->irqstack_sz-sz; ++i)
            dev->irqstack[i] = dev->irqstack[i+sz];
    dev->irqstack_sz -= sz;
    spin_unlock_irqrestore(&dev->irqlock, flags);    
}


static int galil_1800_del_from_siglist(struct galil_1800_state *dev, 
                                       struct task_struct *task)
{
    struct list_head *list;
    struct list_head *head = &dev->siglist;
    struct galil_1800_siglist_t *elem;
    unsigned long flags;

    write_lock_irqsave(&dev->siglock, flags);
    list_for_each(list, head) {
        elem = list_entry(list, struct galil_1800_siglist_t, list);
         if (elem->owner == task)
            break;
    }
    if (list == head) {
        write_unlock_irqrestore(&dev->siglock, flags);
        return -ESRCH;
    }

    if (task->pid != elem->owner->pid) {
    	printk(KERN_WARNING PFX "bad PID (%d) in the signaling process list\n",
                	elem->owner->pid);
        return -EFAULT;
    }
    
    list_del(list);
    write_unlock_irqrestore(&dev->siglock, flags);
    kfree(elem);
    printk(KERN_INFO PFX "deregistered PID %d for device #%d\n",
            task->pid, dev->index);
    return 0;
}

static int galil_1800_add_to_siglist(struct galil_1800_state *dev, 
                                     char * userbuf, struct task_struct *task)
{
    struct list_head *list;
    struct list_head *head = &dev->siglist;
    struct galil_1800_siglist_t *elem;
    unsigned long flags;
    
    if (!userbuf)
        return -EINVAL;

    write_lock_irqsave(&dev->siglock, flags);    
    list_for_each(list, head)
        if (list_entry(list,struct galil_1800_siglist_t,list)->owner==task) { 
            write_unlock_irqrestore(&dev->siglock, flags);
            printk(KERN_INFO PFX "PID %d already registered for device #%d\n",
                    task->pid, dev->index);
            return -EINVAL;
        }
        
    elem = kmalloc(sizeof(struct galil_1800_siglist_t), GFP_ATOMIC);
    if (!elem) {
         printk(KERN_ERR PFX "cannot allocate memory\n");
        return -ENOMEM;   
    }
    elem->owner = task;
    elem->pid = task->pid;
    elem->ubuf = userbuf;
    list_add(&elem->list, head);
    write_unlock_irqrestore(&dev->siglock, flags);
    printk(KERN_INFO PFX "registered PID %d for device #%d\n",
            task->pid, dev->index);
    return 0;
}


static ssize_t galil_1800_read(struct file * filp, char * buf, size_t size, 
                                loff_t *poff)
{
    struct galil_1800_state *dev = 
            (struct galil_1800_state *) filp->private_data;
    char _buf[GALIL_1800_READ_FIFO_SIZE], *p = _buf;
    ssize_t count;
    ssize_t request = (size < GALIL_1800_READ_FIFO_SIZE ? 
                                   size : GALIL_1800_READ_FIFO_SIZE);
    int rc;
    
    for (count = 0; count < request && galil_1800_may_read(dev); ++count)
               *p++ = inb(GALIL_1800_READ_REG(dev));
    if (!count)
        return 0;
    rc = copy_to_user(buf, _buf, count);
    return (rc > 0 ? -EFAULT : count);
}

static ssize_t galil_1800_write(struct file * filp, const char * buf, 
                                size_t size, loff_t *poff)
{
    struct galil_1800_state *dev = 
            (struct galil_1800_state *) filp->private_data;
    char _buf[GALIL_1800_WRITE_FIFO_SIZE], *p = _buf;
    ssize_t count;
    ssize_t request = (size < GALIL_1800_WRITE_FIFO_SIZE ? 
                                   size : GALIL_1800_WRITE_FIFO_SIZE);
    int rc, written = 0;

    if (!request)
        return 0;
    
    if (!buf)
      	return -EINVAL;
    
    rc = copy_from_user(_buf, buf, request);
    if (rc)
        return -EFAULT;
    
    rc = request;
    if (galil_1800_may_write(dev) == 2) {
        
        /* write in burst mode */
        
        if (rc > GALIL_1800_WRITE_FIFO_SIZE / 2) {
            request = GALIL_1800_WRITE_FIFO_SIZE / 2;
            rc -= GALIL_1800_WRITE_FIFO_SIZE / 2;
        } else {
            request = rc;
            rc = 0;
        }
            
        for (count = 0; count < request; ++count, ++p)
            outb_p(*p, GALIL_1800_WRITE_REG(dev));
        written = request;
    }
      
    for (count = 0; count < rc && galil_1800_may_write(dev); ++count, p++)
        outb(*p, GALIL_1800_WRITE_REG(dev));
    return count+written;
}

static int galil_1800_open(struct inode * inode, struct file * file)
{
    int minor = MINOR(inode->i_rdev);
    DECLARE_WAITQUEUE(wait, current);
    struct list_head *list;
    struct galil_1800_state *s;

    read_lock(&galil_1800_devs_lock);
    list_for_each(list, &galil_1800_devs) {
        s = list_entry(list, struct galil_1800_state, devs);
        if (s->index == minor)
            break;
    }
    read_unlock(&galil_1800_devs_lock);

    if (list == &galil_1800_devs)
        return -ENODEV;
    
    
    file->private_data = s;
    
    /* wait for device to become free */
    down(&s->open_sem);
    while (s->open_mode & file->f_mode) {

        if (file->f_flags & O_NONBLOCK) {
            up(&s->open_sem);
            return -EBUSY;
        }
        add_wait_queue(&s->open_wait, &wait);
        __set_current_state(TASK_INTERRUPTIBLE);
        up(&s->open_sem);
        schedule();
        remove_wait_queue(&s->open_wait, &wait);
        set_current_state(TASK_RUNNING);
        if (signal_pending(current))
            return -ERESTARTSYS;
        down(&s->open_sem);
    }

    s->open_mode |= file->f_mode & (FMODE_READ | FMODE_WRITE);
    
	if (++s->opens == 1) {
	    galil_1800_clear_FIFO(s, 0);
    	galil_1800_enable_IRQ(s);
    }

    up(&s->open_sem);
    
    return 0;
}



static int galil_1800_release(struct inode * inode, struct file * file)
{
    struct galil_1800_state *s = 
    	(struct galil_1800_state *) file->private_data;


    down(&s->open_sem);

    if (!--s->opens) {
	    galil_1800_disable_IRQ(s);
    	galil_1800_clear_FIFO(s, 0);
    }
    
    galil_1800_del_from_siglist(s, current);

    s->open_mode &= (~file->f_mode) & (FMODE_READ|FMODE_WRITE);

    up(&s->open_sem);
    
    wake_up(&s->open_wait);
    return 0;
}

static int galil_1800_ioctl(struct inode *inode, struct file *file, 
                            unsigned int cmd, unsigned long arg)
{
    struct galil_1800_state *dev = 
        (struct galil_1800_state *) file->private_data;

    switch (cmd) {
          case GALIL_CLEAR_FIFO:
            galil_1800_clear_FIFO(dev, arg);
            return 0;
            
        case GALIL_MAY_READ:
            return galil_1800_may_read(dev);
            
        case GALIL_MAY_WRITE:
            return galil_1800_may_write(dev);
            
        case GALIL_GET_SERIAL_NUMBER:
            {
#define SN_BUF_SIZE 256
                char buf[SN_BUF_SIZE];
                int rc = galil_1800_get_serial_number(dev, buf, SN_BUF_SIZE);
                return (rc < 0 ? 
                          rc : copy_to_user((void *)arg, buf, strlen(buf)) ? 
                             -EFAULT : 0);
#undef SN_BUF_SIZE
            }
        case GALIL_SIGNAL_INTERRUPTS:
            if ((char *) arg == NULL)
                return galil_1800_del_from_siglist(dev, current);
            else
                return galil_1800_add_to_siglist(dev, (char *) arg, current);
            break;

        case  GALIL_GET_IRQ_STACK:
           {
               int rc =  copy_to_user((void *) arg, dev->irqstack,
                                GALIL_1800_IRQ_STACK_SZ*sizeof(dev->irqstack[0]));
                if (!rc)
                       return -EFAULT;
                rc = dev->irqstack_sz;
                dev->irqstack_sz = 0;
                return rc;
           }
           break;

        case GALIL_READ_2ND_FIFO:
            {
                char buf[GALIL_1800_READ_FIFO2_SIZE];
                int n = galil_1800_read_FIFO2(dev, buf, 
                                              GALIL_1800_READ_FIFO2_SIZE);
                if (!n)
                    return 0;
                return (copy_to_user((void *)arg, buf, n) ? -EFAULT : n);
            }
			break;            
    }
    return -EINVAL;
}

/* module/hotplug stuff */

static int __devinit galil_1800_probe(struct pci_dev *pcidev, 
                                      const struct pci_device_id *devid)
{
    struct galil_1800_state *s;
    
    if (pcidev->irq == 0) 
        return -1;

    if (!(s = kmalloc(sizeof(struct galil_1800_state), GFP_KERNEL))) {
        printk(KERN_ERR PFX "alloc of galil_1800_state struct failed\n");
        return -1;
    }
    memset(s, 0, sizeof(struct galil_1800_state));
    
    init_waitqueue_head(&s->open_wait);
    init_MUTEX(&s->open_sem);
    spin_lock_init(&s->hwlock);
    spin_lock_init(&s->irqlock);
    rwlock_init(&s->siglock);
    INIT_LIST_HEAD(&s->siglist);
    tasklet_init(&s->tlet, galil_1800_send_signals, (unsigned long) s);

    s->index = devindex;
    s->dev = pcidev;
    s->io = pci_resource_start(pcidev, 2);
    s->ctrl_reg = s->io + 4;
    s->irq_reg = s->io + 8;
    s->chan2_reg = s->io + 12;
    s->irq = pcidev->irq;
    s->irqstack_sz = 0;
    
	s->opens = 0;
    
    if (!request_region(s->io, pci_resource_len(pcidev,2),
                GALIL_1800_MODULE_NAME)) {
        printk(KERN_ERR PFX "io ports %#lx->%#lx in use\n",
               s->io, s->io + pci_resource_len(pcidev,2)-1);
        goto err_region;
    }
    if (request_irq(s->irq, galil_1800_interrupt, SA_SHIRQ,
            GALIL_1800_MODULE_NAME, s)) {
        printk(KERN_ERR PFX "irq %u in use\n", s->irq);
        goto err_irq;
    }
    printk(KERN_INFO PFX "device #%d, base I/O: 0x%lx, IRQ: %d\n", 
             s->index, s->io, s->irq);
            
    if (pci_enable_device(pcidev))
        goto err_pci;
    
     pci_set_drvdata(pcidev, s);
    /* pcidev->dma_mask = 0xffffffff; */

    /* put it into driver list */

    write_lock(&galil_1800_devs_lock);     
    list_add_tail(&s->devs, &galil_1800_devs);
    write_unlock(&galil_1800_devs_lock);

        devindex++;
    return 0;

 err_pci:
    printk(KERN_ERR PFX "cannot enable PCI device\n");
    free_irq(s->irq, s);
 err_irq:
    release_region(s->io, pci_resource_len(pcidev,2));
 err_region:
    kfree(s);
    return -1;
}

static void __devexit galil_1800_remove(struct pci_dev *dev)
{
    struct galil_1800_state *s = pci_get_drvdata(dev);
    if (!s)
        return;

    write_lock(&galil_1800_devs_lock);
    list_del(&s->devs);
    write_unlock(&galil_1800_devs_lock);
    
    synchronize_irq();
    free_irq(s->irq, s);
    release_region(s->io, pci_resource_len(dev,2));
    kfree(s);
    pci_set_drvdata(dev, NULL);  
}

static struct __devinitdata pci_device_id galil_1800_idtable[] = {
    {  vendor: PLX_PCI_VENDOR_ID, 
       device: PLX_PCI_DEVICE_ID, 
       subvendor: GALIL_SUB_VENDOR, 
       subdevice: GALIL_SUB_DEVICE_1800, 
       class: 0,
       class_mask: 0, 
       driver_data: 0 },
    { 0,  }
};

static struct pci_driver galil_1800_driver  = {
    name: GALIL_1800_MODULE_NAME,
    id_table: galil_1800_idtable,
    probe: galil_1800_probe,
    remove: __devexit_p(galil_1800_remove),
    save_state: NULL,
    suspend: NULL,
    resume: NULL,
    enable_wake: NULL
};

static struct file_operations galil_1800_fops = {
  read:    galil_1800_read,
  write:   galil_1800_write,
  ioctl:   galil_1800_ioctl,
  open:    galil_1800_open,
  release: galil_1800_release
};

static int __init galil_1800_init(void)
{
    int rc, maj;

    rc = pci_module_init(&galil_1800_driver);

    if (rc == -ENODEV) {
        printk(KERN_INFO PFX "no Galil 1800 device was found\n");
        return rc;
    }

    if (rc) {
        printk(KERN_ERR PFX "cannot register the device driver\n");
        return rc;
    }
    
    maj = (major != -1 ? major : 0);
    rc = register_chrdev(maj, GALIL_1800_MODULE_NAME, &galil_1800_fops);
    if (rc < 0) {
        printk(KERN_ERR PFX "cannot register the device files\n");
        pci_unregister_driver(&galil_1800_driver);
        return 1;
    }
    major = (rc == 0 ? maj : rc);
    SET_MODULE_OWNER(&galil_1800_fops);
    printk(KERN_INFO PFX "module loaded\n");
    return 0;
}

static void __exit galil_1800_cleanup(void)
{
    unregister_chrdev(major, GALIL_1800_MODULE_NAME);
    pci_unregister_driver(&galil_1800_driver);
    printk(KERN_INFO PFX "module unloaded\n");        
}

/* __setup("galil_1800=", galil_1800_setup); */

module_init(galil_1800_init);
module_exit(galil_1800_cleanup);

MODULE_AUTHOR("Shell Technologies s.r.l. (M. Cesati)");
MODULE_DESCRIPTION("Galil 1800 Motion Controller Driver");
MODULE_LICENSE("GPL");
MODULE_PARM(major,"i");
MODULE_PARM_DESC(major, "Major number of the device file");
EXPORT_NO_SYMBOLS;
